"""
This translates from the Tweet Source and Twitter v2 types

Into ViewModel types such as FeedItem

And the rest of the Taxonomy.

"""

from dataclasses import asdict
import os
from flask import session, g, request
import time
import json

from twitter_v2.api import ApiV2TweetSource, TwitterApiV2SocialGraph, ApiV2ConversationSource

from hogumathi_app.view_model import CollectionPage, cleandict
from .view_model import tweet_model_dc_vm, user_model_dc

from hogumathi_app.content_system import register_content_source, get_content, register_hook

DATA_DIR='.data'

def get_tweet_item (tweet_id, me=None):
    
    
    if me:
        twitter_user = session.get(me)
        token = twitter_user['access_token']
    else:
        token = os.environ.get('BEARER_TOKEN')
    
    tweet_source = ApiV2TweetSource(token)
    tweets_response = tweet_source.get_tweet(tweet_id, return_dataclass=True)

    
    #print(response_json)
    if tweets_response.errors:
        # types:
        # https://api.twitter.com/2/problems/not-authorized-for-resource (blocked or suspended)
        # https://api.twitter.com/2/problems/resource-not-found (deleted)
        #print(response_json.get('errors'))
        for err in tweets_response.errors:
            if not 'type' in err:
                print('unknown error type: ' + str(err))
            elif err['type'] == 'https://api.twitter.com/2/problems/not-authorized-for-resource':
                print('blocked or suspended tweet: ' + err['value'])
            elif err['type'] == 'https://api.twitter.com/2/problems/resource-not-found':
                print('deleted tweet: ' + err['value'])
            else:
                print('unknown error')
            
            print(json.dumps(err, indent=2))
    
    
    includes = tweets_response.includes
    tweets = list(map(lambda t: tweet_model_dc_vm(includes, t, me), tweets_response.data))
    
    collection_page = CollectionPage(
        id = tweet_id,
        items = tweets,
        next_token = None # Fixme
        )
    
    return collection_page


def get_bookmarks_feed (user_id, pagination_token=None, max_results=10, me=None):
    
    if not me:
        me = g.get('me') or request.args.get('me')
        
    
    print(f'get_bookmarks_feed. me={me}')
    
    twitter_user = session.get( me )
    
    if not twitter_user:
        return None
        
    token = twitter_user['access_token']
    
    tweet_source = ApiV2TweetSource(token)
    response_tweets = tweet_source.get_bookmarks(user_id,
                        pagination_token = pagination_token,
                        return_dataclass=True,
                        max_results=max_results
                        )
    
    #print(response_json)
    
    includes = response_tweets.includes
    tweets = list(map(lambda t: tweet_model_dc_vm(includes, t, me), response_tweets.data))
    next_token = response_tweets.meta.next_token
    
    query = {}
    
    if next_token:
        query = {
            **query,

        }
    
    user = {
            'id': user_id
        }
    
    ts = int(time.time() * 1000)
    with open(f'{DATA_DIR}/cache/bookmarks_{user_id}_{ts}_{pagination_token}.json', 'wt') as f:
        f.write(json.dumps(cleandict(asdict(response_tweets))))
    
    collection_page = CollectionPage(
        id = user_id, # FIXME this should perhaps be the unresolved id
        items = tweets,
        next_token = next_token
    )
    
    return collection_page

def get_user_feed (user_id, pagination_token=None, me=None, exclude_replies=False, exclude_retweets=True, format=None):
    
    if not me and 'me' in g:
        me = g.me
    
    if 'twitter_user' in g and g.twitter_user:
        token = g.twitter_user['access_token']
        # issue: retweets don't come back if we request non_public_metrics
        is_me = False and user_id == g.twitter_user['id']
    else:
        token = os.environ.get('BEARER_TOKEN')
        is_me = False
    
    
    tweet_source = ApiV2TweetSource(token)
    tweets_response = tweet_source.get_user_timeline(user_id,
                        exclude_replies = exclude_replies,
                        exclude_retweets = exclude_retweets,
                        pagination_token = pagination_token,
                        non_public_metrics = False,
                        return_dataclass=True)
    
    tweets = None
    if not tweets_response:
        print('no response_json')
        
    if tweets_response.meta and tweets_response.meta.result_count == 0:
        print('no results')
        print(tweets_response)
    
    if not tweets_response.includes:
        print(tweets_response)
        print('no tweets_response.includes')
    
    if tweets_response.errors:
        print('profile get_user_timeline errors:')
        print(tweets_response.errors)
    
    ts = int(time.time() * 1000)
    with open(f'{DATA_DIR}/cache/tl_{user_id}_{ts}_{pagination_token}.json', 'wt') as f:
        f.write(json.dumps(cleandict(asdict(tweets_response))))
    
    if tweets_response.data:
        tweets = list(map(lambda t: tweet_model_dc_vm(tweets_response.includes, t, me), tweets_response.data))
    
    next_token = tweets_response.meta.next_token
    
    collection_page = CollectionPage(
        id = user_id,
        items = tweets,
        next_token = next_token
    )
    
    return collection_page

def get_tweets_collection (content_ids, pagination_token=None, max_results=None):
    """
    We might be able to have a generalizer in the content system as well...
    If a source exposes a get many interface then use it. We want to avoid many singular fetches.
    """
    return []
    
def get_user (user_id, me=None):
    
    if me:
        twitter_user = session.get(me)
        token = twitter_user['access_token']
    else:
        token = os.environ.get('BEARER_TOKEN')
    
    social_graph = TwitterApiV2SocialGraph(token)
    users_response = social_graph.get_user(user_id, return_dataclass=True)
    
    print(users_response)
    
    if not len(users_response.data):
        return
    
    user = user_model_dc(users_response.data[0])
    
    return user
    

def get_home_feed (user_id, me, **query_kwargs):
    
    twitter_user = session.get(me)
    token = twitter_user['access_token']
    
    tweet_source = ApiV2TweetSource(token)
    response = tweet_source.get_home_timeline(user_id, **query_kwargs)
    
    #print(json.dumps(response_json, indent=2))
    
    includes = response.includes
    tweets = list(map(lambda t: tweet_model_dc_vm(includes, t, me), response.data))
    next_token = response.meta.next_token
    
    collection_page = CollectionPage(
        id = user_id,
        items = tweets,
        next_token = next_token
    )
    
    return collection_page



def get_tweet_replies (conversation_id, in_reply_to_id=None, pagination_token=None, max_results=None, author_id=None):
    tweet_source = ApiV2TweetSource(token)
    
    only_replies = view == 'replies'
    
    
    tweets = []

    
    skip_embed_replies = False
    
    if view == 'replies':
        replies_response = tweet_source.get_thread(in_reply_to_id,
                                                only_replies=True,
                                                pagination_token = pagination_token,
                                                return_dataclass=True)
    elif view == 'thread':
        skip_embed_replies = True
        replies_response = tweet_source.get_thread(conversation_id,
                                                only_replies=False,
                                                author_id=author_id,
                                                pagination_token = pagination_token,
                                                return_dataclass=True)
                                                
    elif view == 'conversation':
        replies_response = tweet_source.get_thread(conversation_id,
                                                only_replies=False,
                                                pagination_token = pagination_token,
                                                return_dataclass=True)
    elif view == 'tweet':
        replies_response = None
    
    next_token = None
    
    #print("conversation meta:")
    #print(json.dumps(tweets_response.get('meta'), indent=2))
    
    if replies_response and replies_response.meta and replies_response.meta.result_count:
    
        includes = replies_response.includes
        tweets = list(map(lambda t: tweet_model_dc_vm(includes, t, g.me, expand_path=request.args.get('expand'), reply_depth=1), replies_response.data)) + tweets
        
        next_token = replies_response.meta.next_token
    
    # this method is OK except it doesn't work if there are no replies.
    #tweets.append(tweet_model(includes, list(filter(lambda t: t['id'] == tweet_id, includes.get('tweets')))[0], me))

    
    #related_tweets = [] # derived from includes
    
    tweets.reverse()
    
    
    
    query = {}
    
    if next_token:
        query = {
            **query,
            # FIXME only_replies
            'next_data_url': url_for('.get_tweet2_html', tweet_id=tweet_id, pagination_token=next_token, only_replies = '1' if only_replies else '0', author_id = tweets[0].author_id),
            'next_page_url': url_for('.get_tweet2_html', tweet_id=tweet_id, view=view, pagination_token=next_token)
        }
        
    user = {

    }

    
    
    if view == 'replies':
        tweet = tweets[0]
        
        if tweet.id == '1608510741941989378':
            unreplied = [
                UnrepliedSection(
                    description = "Not clear what GS is still.",
                    span = (40, 80)
                )
            ]
            tweet = replace(tweet,
                unreplied = unreplied
                )
        
        expand_parts = request.args.get('expand')
        if expand_parts:
            expand_parts = expand_parts.split(',')
        
        def reply_to_thread_item (fi):
            nonlocal expand_parts
            
            if fi.id == '1609714342211244038':
                print(f'reply_to_thread_item id={fi.id}')
                unreplied = [
                    UnrepliedSection(
                        description = "Is there proof of this claim?",
                        span = (40, 80)
                    )
                ]
                fi = replace(fi,
                    unreplied = unreplied
                    )
            
            children = None
            
            if expand_parts and len(expand_parts) and fi.id == expand_parts[0]:
                expand_parts = expand_parts[1:]
                
                print(f'getting expanded replied for tweet={fi.id}')
                
                expanded_replies_response = tweet_source.get_thread(fi.id,
                                            only_replies=True,
                                            return_dataclass=True)
                if expanded_replies_response.data:
                    print('we got expanded responses data')
                    
                    children =  list(map(lambda t: tweet_model_dc_vm(expanded_replies_response.includes, t, g.me, expand_path=request.args.get('expand'), reply_depth=1), expanded_replies_response.data))
                    children = list(map(reply_to_thread_item, children))
            
            
            return ThreadItem(feed_item=fi, children=children)
            
        children = list(map(reply_to_thread_item, tweets[1:]))
        
        root = ThreadItem(
            feed_item = tweet,
            children = children
        )
        return render_template('tweet-thread.html', user = user, root = root, query = query, page_nav=page_nav, skip_embed_replies=skip_embed_replies, opengraph_info=opengraph_info)
    else:
        return render_template(f'tweet-collection{theme_variant}.html', user = user, tweets = tweets, query = query, page_nav=page_nav, skip_embed_replies=skip_embed_replies, opengraph_info=opengraph_info)


def get_following_users (user_id, me=None, max_results=1000, pagination_token=None):
    
    if me:
        twitter_user = session.get(me)
        token = twitter_user['access_token']
    else:
        token = os.environ.get('BEARER_TOKEN')
    
    social_source = TwitterApiV2SocialGraph(token)
    
    following_resp = social_source.get_following(user_id, 
        max_results=max_results, pagination_token=pagination_token, return_dataclass=True)
    
    ts = int(time.time() * 1000)
    with open(f'{DATA_DIR}/cache/following_{user_id}_{ts}.json', 'wt') as f:
        f.write(json.dumps(cleandict(asdict(following_resp))))
    
    #print(following_resp)
    #run_script('on_user_seen', {'twitter_user': g.twitter_user, 'users': following_resp})
    
    #following = list(map(lambda f: f['id'], following_resp.get('data')))

    following = list(map(user_model_dc, following_resp.data))
    total_count = following_resp.meta.get('result_count')
    next_token = following_resp.meta.get('next_token')
    
    collection_page = CollectionPage(
        id = user_id,
        items = following,
        total_count = total_count,
        next_token = next_token
        )
    
    return collection_page


def get_followers_user (user_id, me=None, max_results=1000, pagination_token=None):
    
    if me:
        twitter_user = session.get(me)
        token = twitter_user['access_token']
    else:
        token = os.environ.get('BEARER_TOKEN')
    
    use_cache = False # this concept is broken for now
    
    
    
    if use_cache: # this concept is broken for now
        print(f'using cache for user {user_id}: {use_cache}')
        with open(f'.data/cache/followers_{user_id}_{pagination_token}_{use_cache}.json', 'rt') as f:
            response_json = json.load(f)
    else:
        social_source = TwitterApiV2SocialGraph(token)
        followers_resp = social_source.get_followers(user_id, max_results=max_results, pagination_token=pagination_token, return_dataclass=True)
    
        ts = int(time.time() * 1000)
        
        print(f'followers cache for {user_id}: {ts}')
        
        with open(f'{DATA_DIR}/cache/followers_{user_id}_{ts}.json', 'wt') as f:
            json.dump(cleandict(asdict(followers_resp)), f, indent=2)
        
        #print(followers_resp)
    #run_script('on_user_seen', {'twitter_user': g.twitter_user, 'users': followers_resp})
    
    #followers = list(map(lambda f: f['id'], followers_resp.get('data')))
    followers = followers_resp.data
    
    followers = list(map(user_model_dc, followers))
    
    followers = list(map(user_model_dc, followers_resp.data))
    total_count = followers_resp.meta.get('result_count')
    next_token = followers_resp.meta.get('next_token')
    
    collection_page = CollectionPage(
        id = user_id,
        items = followers,
        total_count = total_count,
        next_token = next_token
        )
    
    return collection_page

def register_content_sources ():
    register_content_source('twitter:tweets', get_tweets_collection, id_pattern='')
    register_content_source('twitter:tweet:', get_tweet_item, id_pattern='(?P<tweet_id>\d+)')
    register_content_source('twitter:bookmarks:', get_bookmarks_feed, id_pattern='(?P<user_id>\d+)')
    register_content_source('twitter:feed:user:', get_user_feed, id_pattern='(?P<user_id>\d+)')
    register_content_source('twitter:user:', get_user, id_pattern='(?P<user_id>\d+)')
    
    
    register_content_source('twitter:feed:reverse_chronological:user:', get_home_feed, id_pattern='(?P<user_id>\d+)')
    
    
    register_content_source('twitter:tweets:replies:', get_tweet_replies, id_pattern='(?P<conversation_id>\d+)')
    
    register_content_source('twitter:following:users:', get_following_users, id_pattern='(?P<user_id>\d+)')
    register_content_source('twitter:followers:user:', get_followers_user, id_pattern='(?P<user_id>\d+)')
